/***************************************************************************
                          ai.c  -  description
                             -------------------
    begin                : Thu Apr 11 2002
    copyright            : (C) 2001 by Michael Speck
    email                : kulkanie@gmx.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "lgeneral.h"
#include "unit.h"
#include "action.h"
#include "map.h"
#include "ai.h"

#define DEBUG_AI

/*
====================================================================
Externals
====================================================================
*/
extern Player *cur_player;
extern List *units, *avail_units;
extern int map_w, map_h;
extern Map_Tile **map;
extern Mask_Tile **mask;
extern int trgt_type_count;

/*
====================================================================
Internal stuff
====================================================================
*/
List *ai_units = 0; /* units that must be processed */
int score_atk_base = 0;
int score_kill = 0;
int score_loss = 0; /* used to evaluate an attack */
int score_move = 0; /* movement base */
int score_df   = 0; /* defensive fire support bonus */
int score_min_entr = 0;
int score_ini = 0;  /* terrain bonus */
int score_mount = 0; /* malus if unit must mount */
int score_debark_sea = 0; /* bonus for unit near debark tile */
int score_pos = 0; /* nearance to an objective */
int score_rugged = 0; /* malus per point of rugged defense chance */
int score_flag = 0; /* score for conquering a flag */
int score_obj = 0;  /* score if flag was an objective */

/*
====================================================================
Locals
====================================================================
*/

/*
====================================================================
Check if unit has low ammo or fuel.
====================================================================
*/
static int unit_low_fuel( Unit *unit )
{
    if ( !unit_check_fuel_usage( unit ) )
        return 0;
    if ( unit->sel_prop->flags & FLYING ) {
        if ( unit->cur_fuel <= 20 )
            return 1;
        return 0;
    }
    if ( unit->cur_fuel <= 10 )
        return 1;
    return 0;
}
static int unit_low_ammo( Unit *unit )
{
    if ( unit->embark == EMBARK_NONE && unit->cur_ammo <= 2 ) return 1;
    return 0;
}

/*
====================================================================
Get the supply level that is needed to get the wanted absolute
values where ammo or fuel 0 means not to check this value.
====================================================================
*/
#ifdef _1
static int unit_get_supply_level( Unit *unit, int abs_ammo, int abs_fuel )
{
    int ammo_level = 0, fuel_level = 0, miss_ammo, miss_fuel;
    unit_check_supply( unit, UNIT_SUPPLY_ANYTHING, &miss_ammo, &miss_fuel );
    if ( abs_ammo > 0 && miss_ammo > 0 ) {
        if ( miss_ammo > abs_ammo ) miss_ammo = abs_ammo;
        ammo_level = 100 * unit->prop.ammo / miss_ammo;
    }
    if ( abs_fuel > 0 && miss_fuel > 0 ) {
        if ( miss_fuel > abs_fuel ) miss_fuel = abs_fuel;
        ammo_level = 100 * unit->prop.fuel / miss_fuel;
    }
    if ( fuel_level > ammo_level ) 
        return fuel_level;
    return ammo_level;
}

/*
====================================================================
Evaluate a unit's worth and return as single value.
This is a sum up of attack, defense, initiative modified by
experience.
====================================================================
*/
static int unit_evaluate( Unit *unit )
{
    int score = 0;
    int i;
    for ( i = 0; i < trgt_type_count; i++ )
        if ( unit->prop.atks[i] > 0 )
            score += unit->prop.atks[i];
        else
            score += unit->prop.atks[i] / 2;
    score += unit->str - unit->turn_suppr;
    score += unit->prop.ini * 2;
    score += unit->prop.def_grnd;
    score += unit->prop.def_air;
    score += unit->prop.def_cls;
    if ( unit->prop.flags & ARTILLERY || unit->prop.flags & AIR_DEFENSE )
        score += 20;
    score += score * unit->exp_level / 10;
    return score;
}
#endif

/*
====================================================================
Evaluate a unit's attack against target.
  score_base: basic score for attacking
  score_rugged: score added for each rugged def point (0-10)
                of target
  score_kill: score unit receives for each (expected) point of
              strength damage done to target
  score_loss: score that is substracted per strength point
              unit is expected to loose
The final score is stored to 'result' and True if returned if the
attack may be performed else False.
====================================================================
*/
static int unit_evaluate_attack( Unit *unit, Unit *target, int score_base, int score_rugged, int score_kill, int score_loss, int *result )
{
    int unit_dam = 0, target_dam = 0, rugged_def = 0;
    if ( !unit_check_attack( unit, target, UNIT_ACTIVE_ATTACK ) ) return 0;
    unit_get_expected_losses( unit, target, &unit_dam, &target_dam );
    if ( unit_check_rugged_def( unit, target ) )
        rugged_def = unit_get_rugged_def_chance( unit, target );
    if ( rugged_def < 0 ) rugged_def = 0;
    *result = score_base + rugged_def * score_rugged + target_dam * score_kill + unit_dam * score_loss;
    return 1;
}

/*
====================================================================
Get the distance of 'unit' and position of object of a special
type.
====================================================================
*/
enum {
    AI_FIND_DEPOT,
    AI_FIND_OWN_TOWN,
    AI_FIND_ENEMY_TOWN,
    AI_FIND_OWN_OBJ,
    AI_FIND_ENEMY_OBJ
};
static int ai_check_hex_type( Unit *unit, int type, int x, int y )
{
    switch ( type ) {
        case AI_FIND_DEPOT:
            if ( map_is_allied_depot( &map[x][y], unit ) )
                return 1;
            break;
        case AI_FIND_ENEMY_OBJ:
            if ( !map[x][y].obj ) return 0;
        case AI_FIND_ENEMY_TOWN:
            if ( map[x][y].player && !player_is_ally( unit->player, map[x][y].player ) )
                return 1;
            break;
        case AI_FIND_OWN_OBJ:
            if ( !map[x][y].obj ) return 0;
        case AI_FIND_OWN_TOWN:
            if ( map[x][y].player && player_is_ally( unit->player, map[x][y].player ) )
                return 1;
            break;
    }
    return 0;
}
static int ai_get_dist( Unit *unit, int x, int y, int type, int *dx, int *dy, int *dist )
{
    int found = 0, length;
    int i, j;
    *dist = 999;
    for ( i = 0; i < map_w; i++ )
        for ( j = 0; j < map_h; j++ )
            if ( ai_check_hex_type( unit, type, i, j ) ) {
                length = get_dist( i, j, x, y );
                if ( *dist > length ) {
                    *dist = length;
                    *dx = i; *dy = j;
                    found = 1;
                }
            }
    return found;
}

/*
====================================================================
Approximate destination by best move position in range.
====================================================================
*/
static int ai_approximate( Unit *unit, int dx, int dy, int *x, int *y )
{
    int i, j, dist = get_dist( unit->x, unit->y, dx, dy ) + 1;
    *x = unit->x; *y = unit->y;
    map_clear_mask( F_AUX );
    map_get_unit_move_mask( unit );
    for ( i = 0; i < map_w; i++ )
        for ( j = 0; j < map_h; j++ )
            if ( mask[i][j].in_range && !mask[i][j].blocked )
                mask[i][j].aux = get_dist( i, j, dx, dy ) + 1;
    for ( i = 0; i < map_w; i++ )
        for ( j = 0; j < map_h; j++ )
            if ( dist > mask[i][j].aux && mask[i][j].aux > 0 ) {
                dist = mask[i][j].aux;
                *x = i; *y = j;
            }
    return ( *x != unit->x || *y != unit->y );
}

/*
====================================================================
Get the best target for unit if any.
====================================================================
*/
static int ai_get_best_target( Unit *unit, int x, int y, Unit **target, int *score )
{
    int old_x = unit->x, old_y = unit->y;
    Unit_Lib_Entry *old_sel_prop = unit->sel_prop;
    int i, j;
    int result;
    Unit *entry;
    List *targets = list_create( LIST_NO_AUTO_DELETE, LIST_NO_CALLBACK );
    unit->x= x; unit->y = y;
    /* use trsp_prop if needed */
    if ( mask[x][y].mount && unit->trsp_prop.id )
        unit->sel_prop = &unit->trsp_prop;
    /* gather targets */
    for ( i = x - unit->sel_prop->rng - 1; i <= x + unit->sel_prop->rng + 1; i++ )
        for ( j = y - unit->sel_prop->rng - 1; j <= y + unit->sel_prop->rng + 1; j++ ) {
            if ( i <= 0 || j <= 0 || i >= map_w - 1 || j >= map_h - 1 )
                continue;
            if ( !mask[i][j].spot ) 
                continue;
            if ( map[i][j].a_unit && unit_check_attack( unit, map[i][j].a_unit, UNIT_ACTIVE_ATTACK ) )
                list_add( targets, map[i][j].a_unit );
            if ( map[i][j].g_unit && unit_check_attack( unit, map[i][j].g_unit, UNIT_ACTIVE_ATTACK ) )
                list_add( targets, map[i][j].g_unit );
        }
    /* get best target */
    list_reset( targets ); *target = 0; *score = 0;
    while ( ( entry = list_next( targets ) ) )
        if ( unit_evaluate_attack( unit, entry, score_atk_base, score_rugged, score_kill, score_loss, &result ) ) {
            if ( result > *score ) {
                *target = entry;
                *score = result;
            }
        }
    list_delete( targets );
    unit->x = old_x; unit->y = old_y;
    unit->sel_prop = old_sel_prop;
    return (*target) != 0;
}

/*
====================================================================
Handle unit and return position and target (if any) where to
move unit. Same position means not to move the unit.
====================================================================
*/
static void ai_evaluate_hex( Unit *unit, int x, int y, Unit **target, int *score )
{
    int ox, oy, odist;
    int result = 0, i, j, nx, ny;
    *score = 0;
    *target = 0;
    /* base */
    if ( unit->x != x || unit->y != y )
        *score += score_move;
    /* terrain */
    if ( !(unit->sel_prop->flags & FLYING ) ) {
        //*score += map[x][y].terrain->max_ini * score_ini;
        *score += map[x][y].terrain->min_entr * score_min_entr;
    }
    /* mount */
    if ( mask[x][y].mount )
        *score += score_mount;
    /* attack */
    if ( !( unit->sel_prop->flags & ATTACK_FIRST ) || ( unit->x == x && unit->y == y ) )
        if ( ai_get_best_target( unit, x, y, target, &result ) )
            *score += result;
    /* conquer flag? */
    if ( !(unit->sel_prop->flags & FLYING ) )
        if ( map[x][y].player && !player_is_ally( unit->player, map[x][y].player ) )
            if ( map[x][y].g_unit == 0 ) {
                *score += score_flag;
                if ( map[x][y].obj )
                    *score += score_obj;
            }
    /* debark bonus */
    if ( unit->embark == EMBARK_SEA ) {
        if ( map_check_unit_debark( unit, x, y, EMBARK_SEA ) )
            *score += score_debark_sea * 2;
        else
            for ( i = 0; i < 6; i++ )
                if ( get_close_hex_pos( x, y, i, &nx, &ny ) )
                    if ( map_check_unit_debark( unit, nx, ny, EMBARK_SEA ) ) {
                        *score += score_debark_sea;
                        break;
                    }
    }
    /* position */
    if ( !(unit->sel_prop->flags & FLYING ) ) {
        if ( cur_player->strat > 0 ) {
            if ( ai_get_dist( unit, x, y, AI_FIND_ENEMY_OBJ, &ox, &oy, &odist ) )
                *score += odist * score_pos;
        }
        else
            if ( cur_player->strat < 0 ) {
                if ( ai_get_dist( unit, x, y, AI_FIND_OWN_OBJ, &ox, &oy, &odist ) )
                    *score += odist * score_pos;
            }
    }
    /* defensive support */
    for ( i = x - 3; i <= x + 3; i++ )
        for ( j = y - 3; j <= y + 3; j++ ) {
            if ( i <= 0 || j <= 0 || i >= map_w - 1 || j >= map_h - 1 )
                continue;
            if ( map[i][j].g_unit && player_is_ally( unit->player, map[i][j].g_unit->player ) ) {
                if ( map[i][j].g_unit->sel_prop->flags & ARTILLERY )
                    if ( !(unit->sel_prop->flags & FLYING ) )
                        *score += score_df;
                if ( map[i][j].g_unit->sel_prop->flags & AIR_DEFENSE )
                    if ( unit->sel_prop->flags & FLYING )
                        *score += score_df;
            }
        }
#ifdef DEBUG_AI
//    printf( "  eval: %2i,%2i: %2i\n", x, y, *score );
#endif
}
typedef struct {
    int x, y;
} AI_Pos;
static AI_Pos *ai_create_pos( int x, int y )
{
    AI_Pos *pos = calloc( 1, sizeof( AI_Pos ) );
    pos->x = x; pos->y = y;
    return pos;
}
static void ai_handle_unit( Unit *unit, int *x, int *y, Unit **target )
{
    Unit *new_target;
    int nx, ny, i;
    List *list = list_create( LIST_NO_AUTO_DELETE, LIST_NO_CALLBACK );
    AI_Pos *pos;
    int score = -10000, result;
    /* get move mask */
    map_get_unit_move_mask( unit );
    *x = unit->x; *y = unit->y; *target = 0;
    /* evaluate all positions */
    list_add( list, ai_create_pos( unit->x, unit->y ) );
    while ( list->count > 0 ) {
        pos = list_dequeue( list );
        ai_evaluate_hex( unit, pos->x, pos->y, &new_target, &result );
        if ( result > score ) {
            score = result;
            *target = new_target;
            *x = pos->x; *y = pos->y;
        }
        for ( i = 0; i < 6; i++ )
            if ( get_close_hex_pos( pos->x, pos->y, i, &nx, &ny ) )
                if ( ( mask[nx][ny].in_range && !mask[nx][ny].blocked ) || mask[nx][ny].sea_embark ) {
                    mask[nx][ny].in_range = 0; 
                    mask[nx][ny].sea_embark = 0;
                    list_add( list, ai_create_pos( nx, ny ) );
                }
        free( pos );
    }
    list_delete( list );
#ifdef DEBUG_AI
//    printf( "  evaluation: %i\n", score );
#endif
}

/*
====================================================================
Exports
====================================================================
*/

/*
====================================================================
Initiate turn
====================================================================
*/
void ai_init( void )
{
    List *list; /* used to speed up the creation of ai_units */
    Unit *unit;
#ifdef DEBUG_AI
    printf( "AI Turn: %s\n", cur_player->name );
#endif
    /* get all cur_player units, those with defensive fire come first */
    list = list_create( LIST_NO_AUTO_DELETE, LIST_NO_CALLBACK );
    list_reset( units );
    while ( ( unit = list_next( units ) ) )
        if ( unit->player == cur_player )
            list_add( list, unit );
    ai_units = list_create( LIST_NO_AUTO_DELETE, LIST_NO_CALLBACK );
    list_reset( list );
    while ( ( unit = list_next( list ) ) )
        if ( unit->sel_prop->flags & ARTILLERY || unit->sel_prop->flags & AIR_DEFENSE ) {
            list_add( ai_units, unit );
            list_delete_item( list, unit );
        }
    list_reset( list );
    while ( ( unit = list_next( list ) ) )
        list_add( ai_units, unit );
    list_delete( list );
    list_reset( ai_units );
    list_reset( avail_units );
    /* set the evaluation scores */
    /* strategy: -2 very defensive
                  2 very aggressive */
    score_atk_base = cur_player->strat * 20;
    score_kill = 20;
    score_loss = ( 2 - cur_player->strat ) * -10;
    score_move = 0;//cur_player->strat * 10;
    score_df   = 20;
    score_ini = 3;
    score_min_entr = 3;
    score_mount = -20;
    score_debark_sea = 300;
    score_pos = -10;
    score_rugged = -10;
    score_flag = 100;
    score_obj = 100;
#ifdef DEBUG_AI
    printf( "Units: %i\n", ai_units->count );
    printf( "AI Initialized\n" );
#endif
}

/*
====================================================================
Queue next actions (if these actions were handled by the engine
this function is called again and again until the end_turn
action is received).
====================================================================
*/
void ai_run( void )
{
    int i, j, x, y, dx, dy, dist, found;
    Unit *target;
    Unit *unit = 0;
    /* deploy unit? */
    if ( avail_units->count > 0 && ( unit = list_next( avail_units ) ) ) {
        map_clear_mask( F_AUX );
        for ( i = 0; i < map_w; i++ )
            for ( j = 0; j < map_h; j++ )
                if ( map_check_deploy( unit, i, j ) )
                    if ( ai_get_dist( unit, i, j, AI_FIND_ENEMY_OBJ, &x, &y, &dist ) )
                        mask[i][j].aux = dist + 1;
        dist = 1000; found = 0;
        for ( i = 0; i < map_w; i++ )
            for ( j = 0; j < map_h; j++ )
                if ( mask[i][j].aux > 0 && mask[i][j].aux < dist ) {
                    dist = mask[i][j].aux;
                    x = i; y = j;
                    found = 1;
                }
        if ( found ) {
            action_queue_deploy( unit, x, y );
            list_reset( avail_units );
            list_add( ai_units, unit );
#ifdef DEBUG_AI
    printf( "Deploy: %s (%i,%i)\n", unit->name, unit->x, unit->y );
#endif
            return;
        }
    }
    /* get next unit */
    unit = list_next( ai_units );
    /* end turn? */
    if ( unit == 0 ) {
        action_queue_end_turn();
        return;
    }
#ifdef DEBUG_AI
    printf( "%s:\n", unit->name );
#endif
    /* supply */
    if ( ( unit_low_fuel( unit ) || unit_low_ammo( unit ) ) ) {
        if ( unit->supply_level > 0 ) {
            action_queue_supply( unit );
#ifdef DEBUG_AI
            printf( "  supply\n" );
#endif
            return;
        }
        else {
#ifdef DEBUG_AI
            printf( "  searching depot\n" );
#endif
            if ( unit->sel_prop->flags & FLYING ) {
                if ( ai_get_dist( unit, unit->x, unit->y, AI_FIND_DEPOT,
                                  &dx, &dy, &dist ) )
                    if ( ai_approximate( unit, dx, dy, &x, &y ) ) {
                        action_queue_move( unit, x, y );
#ifdef DEBUG_AI
                        printf( "  move to %i,%i\n", x, y );
#endif
                        return;
                    }
            }
            else
                if ( ai_get_dist( unit, unit->x, unit->y, AI_FIND_DEPOT,
                                  &dx, &dy, &dist ) )
                    if ( ai_approximate( unit, dx, dy, &x, &y ) ) {
                        action_queue_move( unit, x, y );
#ifdef DEBUG_AI
                        printf( "  move to %i,%i\n", x, y );
#endif
                        return;
                    }
        }
    }
    /* move unit */
    ai_handle_unit( unit, &x, &y, &target );
    if ( unit->x != x || unit->y != y ) {
        if ( map_check_unit_debark( unit, x, y, EMBARK_SEA ) ) {
            action_queue_debark_sea( unit, x, y );
#ifdef DEBUG_AI
            printf( "  debark at %i,%i\n", x, y );
#endif
        }
        else {
            action_queue_move( unit, x, y );
#ifdef DEBUG_AI
            printf( "  move to %i,%i\n", x, y );
#endif
        }
    }
    if ( target ) {
        action_queue_attack( unit, target );
#ifdef DEBUG_AI
        printf( "  attack %s\n", target->name );
#endif
    }
}

/*
====================================================================
Undo the steps (e.g. memory allocation) made in ai_init()
====================================================================
*/
void ai_finalize( void )
{
    if ( ai_units ) {
        list_delete( ai_units );
        ai_units = 0;
    }
#ifdef DEBUG_AI
    printf( "AI Finalized\n" );
#endif
}

